<?php
/**
 * SeekQuarry/Yioop --
 * Open Source Pure PHP Search Engine, Crawler, and Indexer
 *
 * Copyright (C) 2009 - 2015  Chris Pollett chris@pollett.org
 *
 * LICENSE:
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * END LICENSE
 *
 * @author Chris Pollett chris@pollett.orgs
 * @license http://www.gnu.org/licenses/ GPL3
 * @link http://www.seekquarry.com/
 * @copyright 2009 - 2015
 * @filesource
 */
namespace seekquarry\yioop\models;

use seekquarry\yioop\configs as C;
use seekquarry\yioop\library as L;
use seekquarry\yioop\library\CrawlConstants;
use seekquarry\yioop\library\FetchUrl;
use seekquarry\yioop\library\IndexShard;
use seekquarry\yioop\library\PhraseParser;
use seekquarry\yioop\library\UrlParser;

/** For getLocaleTag*/
require_once __DIR__.'/../library/LocaleFunctions.php';
/**
 * Used to manage data related to video, news, and other search sources
 * Also, used to manage data about available subsearches seen in SearchView
 *
 * @author Chris Pollett
 */
class SourceModel extends ParallelModel
{
    /** Mamimum number of feeds to download in one try */
    const MAX_FEEDS_ONE_GO = 100;
    /**
     * Controls which tables and the names of tables
     * underlie the given model and should be used in a getRows call
     * As SourceModel is used for both media sources and subsearches.
     * The underlying table might be MEDIA_SOURCE or it might be SUBSEARCH.
     * The $args variable is a string which is assumed to say which.
     *
     * @param string $args if is "SUBSEARCH" then the SUBSEARCH table will
     *     be used by getRows rather than MEDIA_SOURCE.
     * @return string which table to use
     */
    public function fromCallback($args = null)
    {
        if ($args == "SUBSEARCH") {
            return "SUBSEARCH";
        }
        return "MEDIA_SOURCE";
    }
    /**
     * Returns a list of media sources such as (video, rss sites) and their
     * URL and thumb url formats, etc
     *
     * @param string $source_type the particular kind of media source to return
     *     for example, video
     * @param bool $has_no_feed_items if true returns only those items which
     *     have not feed_items associated with them.
     * @return array a list of web sites which are either video or news sites
     */
    public function getMediaSources($source_type = "",
        $has_no_feed_items = false)
    {
        $db = $this->db;
        $sources = [];
        $params = [];
        $sql = "SELECT M.* FROM MEDIA_SOURCE M";
        if ($source_type !="") {
            $sql .= " WHERE TYPE=:type";
            $params = [":type" => $source_type];
        }
        if ($has_no_feed_items) {
            if ($source_type == "") {
                $sql .= " WHERE ";
            } else {
                $sql .= " AND ";
            }
            $sql .= " NOT EXISTS
                (SELECT * FROM FEED_ITEM F
                WHERE F.SOURCE_NAME = M.NAME)";
        }
        $i = 0;
        $result = $db->execute($sql, $params);
        while ($sources[$i] = $db->fetchArray($result)) {
            $i++;
        }
        unset($sources[$i]); //last one will be null
        return $sources;
    }
    /**
     * Return the media source by the name of the source
     * @param string $timestamp of the media source to look up
     * @return array associative array with SOURCE_NAME, TYPE, SOURCE_URL,
     *     AUX_INFO, and LANGUAGE
     */
    public function getMediaSource($timestamp)
    {
        $db = $this->db;
        $sql = "SELECT * FROM MEDIA_SOURCE WHERE TIMESTAMP = ?";
        $result = $db->execute($sql, [$timestamp]);
        if (!$result) {
            return false;
        }
        $row = $db->fetchArray($result);
        return $row;
    }
   /**
    * Receives a request to get machine data for an array of hashes of urls
    * @return a list of urls of machines used by this instance of yioop for
    *   crawling
    */
    public function getMachineHashUrls()
    {
        $db = $this->db;
        $machines = [];
        $sql = "SELECT DISTINCT URL FROM MACHINE";
        $result = $db->execute($sql);
        $machine_hashes = [];
        while ($row = $db->fetchArray($result)) {
            $machine_hashes[] = L\crawlHash($row['URL']);
        }
        return $machine_hashes;
    }
    /**
     * Used to add a new video, rss, html news, or other sources to Yioop
     *
     * @param string $name
     * @param string $source_type whether video, rss, etc
     * @param string $source_url url regex of resource (video) or actual
     *     resource (rss). Not quite a real regex you add {} to the
     *     location in the url where the name of the particular video
     *     should go http://www.youtube.com/watch?v={}&
     *     (anything after & is ignored, so between = and & will be matched
     *     as the name of a video)
     * @param string $aux_info regex of where to get thumbnails for videos
     *      based on match of $source_url, for example,
     *      http://img.youtube.com/vi/{}/2.jpg
     *      For html news feeds has xpaths to scrape news items from regex
     * @param string $language the locale tag for the media source (rss)
     */
    public function addMediaSource($name, $source_type, $source_url, $aux_info,
        $language = C\DEFAULT_LOCALE)
    {
        $db = $this->db;
        $sql = "INSERT INTO MEDIA_SOURCE VALUES (?,?,?,?,?,?)";

        $db->execute($sql, [time(), $name, $source_type, $source_url,
            $aux_info, $language]);
    }
    /**
     * Used to update the fields stored in a MEDIA_SOURCE row according to
     * an array holding new values
     *
     * @param array $source_info updated values for a MEDIA_SOURCE row
     */
    public function updateMediaSource($source_info)
    {
        $timestamp = $source_info['TIMESTAMP'];
        unset($source_info['TIMESTAMP']);
        unset($source_info['NAME']);
        $sql = "UPDATE MEDIA_SOURCE SET ";
        $comma ="";
        $params = [];
        foreach ($source_info as $field => $value) {
            $sql .= "$comma $field=? ";
            $comma = ",";
            $params[] = $value;
        }
        $sql .= " WHERE TIMESTAMP=?";
        $params[] = $timestamp;
        $this->db->execute($sql, $params);
    }
    /**
     * Deletes the media source whose id is the given timestamp
     *
     * @param int $timestamp of media source to be deleted
     */
    public function deleteMediaSource($timestamp)
    {
        $sql = "SELECT * FROM MEDIA_SOURCE WHERE TIMESTAMP='$timestamp'";
        $result = $this->db->execute($sql);
        if ($result) {
            $row = $this->db->fetchArray($result);
            if (isset($row['TYPE']) && $row['TYPE'] == "rss") {
                if ($row['NAME'] != "") {
                    $sql = "DELETE FROM FEED_ITEM WHERE SOURCE_NAME=?";
                    $this->db->execute($sql, [$row['NAME']]);
                }
            }
        }
        $sql = "DELETE FROM MEDIA_SOURCE WHERE TIMESTAMP=?";
        $this->db->execute($sql, [$timestamp]);
    }
    /**
     * Returns a list of the subsearches used by the current Yioop instances
     * including their names translated to the current locale
     *
     * @return array associative array containing subsearch info name in locale,
     *    folder name, index, number of results per page
     */
    public function getSubsearches()
    {
        $subsearches = [];
        $db = $this->db;
        $locale_tag = L\getLocaleTag();
        $sql = "SELECT LOCALE_ID FROM LOCALE ".
            "WHERE LOCALE_TAG = ? " . $db->limitOffset(1);
        $result = $db->execute($sql, [$locale_tag]);
        $row = $db->fetchArray($result);
        $locale_id = $row['LOCALE_ID'];
        $sql = "SELECT S.LOCALE_STRING AS LOCALE_STRING, ".
            "S.FOLDER_NAME AS FOLDER_NAME, ".
            " S.PER_PAGE AS PER_PAGE, ".
            " S.INDEX_IDENTIFIER AS INDEX_IDENTIFIER, ".
            " T.TRANSLATION_ID AS TRANSLATION_ID FROM ".
            " SUBSEARCH S, TRANSLATION T WHERE  ".
            " T.IDENTIFIER_STRING = S.LOCALE_STRING";
        $i = 0;
        $result = $db->execute($sql);
        $sub_sql = "SELECT TRANSLATION AS SUBSEARCH_NAME ".
            "FROM TRANSLATION_LOCALE ".
            " WHERE TRANSLATION_ID=? AND LOCALE_ID=? " . $db->limitOffset(1);
            // maybe do left join at some point
        while ($subsearches[$i] = $db->fetchArray($result)) {
            $id = $subsearches[$i]["TRANSLATION_ID"];
            $result_sub =  $db->execute($sub_sql, [$id, $locale_id]);
            $translate = false;
            if ($result_sub) {
                $translate = $db->fetchArray($result_sub);
            }
            if ($translate) {
                $subsearches[$i]['SUBSEARCH_NAME'] =
                    $translate['SUBSEARCH_NAME'];
            } else {
                $subsearches[$i]['SUBSEARCH_NAME'] = $this->translateDb(
                    $subsearches[$i]['LOCALE_STRING'], C\DEFAULT_LOCALE);
            }
            $i++;
        }
        unset($subsearches[$i]); //last one will be null
        return $subsearches;
    }
    /**
     * Return the media source by the name of the source
     * @param string $folder_name
     * @return array
     */
    public function getSubsearch($folder_name)
    {
        $db = $this->db;
        $sql = "SELECT * FROM SUBSEARCH WHERE FOLDER_NAME = ?";
        $result = $db->execute($sql, [$folder_name]);
        if (!$result) {
            return false;
        }
        $row = $db->fetchArray($result);
        return $row;
    }
    /**
     * Given the folder name for a subsearch and a locale tag return the
     * natural language name in that for the subsearch
     *
     * @param string $folder_name of subsearch want to look up
     * @param string $locale_tag of language want human understandable
     *      subsearch name
     * @return string natural language name of subsearch
     */
    public function getSubsearchName($folder_name, $locale_tag)
    {
        $subsearch = $this->getSubsearch($folder_name);
        if(isset($subsearch['LOCALE_STRING'])) {
            return $this->translateDb(
                    $subsearch['LOCALE_STRING'], $locale_tag);
        }
        return $folder_name;
    }
    /**
     * Adds a new subsearch to the list of subsearches. This are displayed
     * at the top od the Yioop search pages.
     *
     * @param string $folder_name name of subsearch in terms of urls
     *     (not translated name that appears in the subsearch bar)
     * @param string $index_identifier timestamp of crawl or mix to be
     *     used for results of subsearch
     * @param int $per_page number of search results per page when this
     *     subsearch is used
     */
    public function addSubsearch($folder_name, $index_identifier, $per_page)
    {
        $db = $this->db;
        $locale_string = "db_subsearch_".$folder_name;
        $sql = "INSERT INTO SUBSEARCH VALUES (?, ?, ?, ?)";
        $db->execute($sql, [$locale_string, $folder_name, $index_identifier,
            $per_page]);
        $sql = "INSERT INTO TRANSLATION VALUES (?, ?)";
        $db->execute($sql, [time(), $locale_string]);
    }
    /**
     * Used to update the fields stored in a SUBSEARCH row according to
     * an array holding new values
     *
     * @param array $search_info updated values for a SUBSEARCH row
     */
    public function updateSubsearch($search_info)
    {
        $folder_name = $search_info['FOLDER_NAME'];
        unset($search_info['FOLDER_NAME']);
        $sql = "UPDATE SUBSEARCH SET ";
        $comma ="";
        $params = [];
        foreach ($search_info as $field => $value) {
            $sql .= "$comma $field=? ";
            $comma = ",";
            $params[] = $value;
        }
        $sql .= " WHERE FOLDER_NAME=?";
        $params[] = $folder_name;
        $this->db->execute($sql, $params);
    }
    /**
     * Deletes a subsearch from the subsearch table and removes its
     * associated translations
     *
     * @param string $folder_name of subsearch to delete
     */
    public function deleteSubsearch($folder_name)
    {
        $db = $this->db;
        $locale_string = "db_subsearch_".$folder_name;
        $sql = "SELECT * FROM TRANSLATION WHERE IDENTIFIER_STRING = ?";
        $result = $db->execute($sql, [$locale_string]);
        if (isset($result)) {
            $row = $db->fetchArray($result);
            if (isset($row["TRANSLATION_ID"])) {
                $sql = "DELETE FROM TRANSLATION_LOCALE WHERE ".
                    "TRANSLATION_ID=?";
                $db->execute($sql, [$row["TRANSLATION_ID"]]);
            }
        }
        $sql = "DELETE FROM SUBSEARCH WHERE FOLDER_NAME=?";
        $db->execute($sql, [$folder_name]);

        $sql = "DELETE FROM TRANSLATION WHERE IDENTIFIER_STRING = ?";
        $db->execute($sql, [$locale_string]);
    }
    /**
     * For each feed source downloads the feeds, checks which items are
     * not in the database, adds them. This method does not update
     * the inverted index shard.
     *
     * @param int $age how many seconds old records should be ignored
     * @param string $media_mode might be one of name_server or distributed --
     *      this controls whether the name server should be contacted to get
     *      a list of news sources (distributed case) or whether this is the
     *      non-distributed case and data should be gotten from local machine
     */
    public function updateFeedItems($age = C\ONE_WEEK,
        $media_mode = "name_server")
    {
        $feeds = $this->getNewsSources(false, $media_mode);
        if (!is_array($feeds)) {
            L\crawlLog("This media updater is NOT responsible for any feeds!");
            return;
        }
        L\crawlLog("This media updater is responsible for the feeds:");
        $i = 1;
        foreach ($feeds as $feed) {
            L\crawlLog("  $i. ".$feed["NAME"]);
            $i++;
        }
        $num_feeds = count($feeds);
        $feeds_one_go = self::MAX_FEEDS_ONE_GO;
        $limit = 0;
        while ($limit < $num_feeds) {
            $feeds_batch = array_slice($feeds, $limit, $feeds_one_go);
            $this->updateFeedItemsOneGo($feeds, $age);
            $limit += $feeds_one_go;
        }
    }
    /**
     * Gets a list of rss news feed info either from the local database or
     * from the name server (in multiple news feeder setting)
     *
     * @param bool $use_cache whether to use a previously cached news list
     * @param string $media_mode might be one of name_server or distributed --
     *      this controls whether the name server should be contacted to get
     *      a list of news sources (distributed case) or whether this is the
     *      non-distributed case and data should be gotten from local machine
     * @return array $feeds info about all the news sources this machine
     *      is responsible for
     */
    public function getNewsSources($use_cache = true,
        $media_mode = "name_server")
    {
        static $feeds = [];
        if ($use_cache && $feeds != []) {
            return $feeds;
        }
        if ($media_mode == "distributed") {
            L\crawlLog("Getting news sources from name server...");
            $current_machine = $this->getCurrentMachine();
            $pre_feeds = $this->execMachines("getNewsSources",
               [C\NAME_SERVER], $current_machine);
            $feeds = [];
            if (isset($pre_feeds[0][self::PAGE])) {
                $feeds = unserialize(L\webdecode($pre_feeds[0][self::PAGE]));
            }
        } else {
            $db = $this->db;
            $sql = "SELECT * FROM MEDIA_SOURCE WHERE (TYPE='rss'
                 OR TYPE='html')";
            $result = $db->execute($sql);
            $i = 0;
            while ($feeds[$i] = $this->db->fetchArray($result)) {
                $aux_parts = explode("###",
                        html_entity_decode($feeds[$i]['AUX_INFO']));
                if ($feeds[$i]['TYPE'] == 'html') {
                    list($feeds[$i]['CHANNEL_PATH'], $feeds[$i]['ITEM_PATH'],
                    $feeds[$i]['TITLE_PATH'], $feeds[$i]['DESCRIPTION_PATH'],
                    $feeds[$i]['LINK_PATH']) =
                        $aux_parts;
                    $offset = 5;
                } elseif ($feeds[$i]['TYPE'] == 'rss') {
                    $offset = 0;
                }
                if (isset($aux_parts[$offset])) {
                    $feeds[$i]['IMAGE_XPATH'] = $aux_parts[$offset];
                } else {
                    $feeds[$i]['IMAGE_XPATH'] = "";
                }
                $i++;
            }
            unset($feeds[$i]); //last one will be null
        }
        return $feeds;
    }
    /**
     * Downloads one batch of $feeds_one_go feed items for @see updateFeedItems
     * For each feed source downloads the feeds, checks which items are
     * not in the database, adds them. This method does not update
     * the inverted index shard.
     *
     * @param array $feeds list of feeds to download
     * @param int $age how many seconds old records should be ignored
     */
    public function updateFeedItemsOneGo($feeds, $age = C\ONE_WEEK)
    {
        $feeds = FetchUrl::getPages($feeds, false, 0, null, "SOURCE_URL",
            CrawlConstants::PAGE, true, null, true);
        $sql = "UPDATE MEDIA_SOURCE SET LANGUAGE=? WHERE TIMESTAMP=?";
        libxml_use_internal_errors(true);
        foreach ($feeds as $feed) {
            if (!isset($feed[CrawlConstants::PAGE]) ||
                !$feed[CrawlConstants::PAGE]) {
                L\crawlLog("...No data in feed skipping.");
                continue;
            }
            L\crawlLog("Updating {$feed['NAME']}. Making dom object from feed.");
            $is_html = ($feed['TYPE'] == 'html') ? true : false;
            $dom = new \DOMDocument();
            //strip namespaces
            $page = preg_replace('@<(/?)(\w+\s*)\:@u', '<$1',
                $feed[CrawlConstants::PAGE]);
            if (isset($feed['IMAGE_XPATH'])) {
                $feed['IMAGE_XPATH'] = preg_replace('@/(\s*\w+\s*)\:@u', '/',
                    $feed['IMAGE_XPATH']);
            }
            if ($is_html) {
                @$dom->loadHTML($page);
            } else {
                /*
                    We parse using loadHTML as less strict. loadHTML
                    auto-closes link tags immediately after open link
                    so to avoid this we replace link with xlink
                 */
                $page = preg_replace("@<link@", "<slink", $page);
                $page = preg_replace("@</link@", "</slink", $page);
                $page = preg_replace("@pubDate@i", "pubdate", $page);
                $page = preg_replace("@&lt;@", "<", $page);
                $page = preg_replace("@&gt;@", ">", $page);
                $page = preg_replace("@<!\[CDATA\[(.+?)\]\]>@", '$1', $page);
                // we also need a hack to make UTF-8 work correctly
                @$dom->loadHTML('<?xml encoding="UTF-8">' . $page);
                foreach ($dom->childNodes as $item)
                if ($item->nodeType == XML_PI_NODE)
                    $dom->removeChild($item);
                $dom->encoding = 'UTF-8';
            }
            L\crawlLog("...done. Extracting info about whole feed.");
            $lang = "";
            if ($feed['TYPE'] != 'html' &&
                !isset($feed["LANGUAGE"]) || $feed["LANGUAGE"] == "") {
                $languages = $dom->getElementsByTagName('language');
                if ($languages && is_object($languages) &&
                    is_object($languages->item(0))) {
                    $lang = $languages->item(0)->textContent;
                    $db->execute($sql, [$lang, $feed['TIMESTAMP']]);
                }
            } elseif (isset($feed["LANGUAGE"]) && $feed["LANGUAGE"] != "") {
                $lang = $feed["LANGUAGE"];
            } else {
                $lang = C\DEFAULT_LOCALE;
            }
            L\crawlLog("...Language is $lang. Getting channel, finding nodes.");
            if ($is_html) {
                $sub_dom = $this->getTags($dom, $feed['CHANNEL_PATH']);
                if (!$sub_dom) {
                    L\crawlLog("... Scraper couldn't parse channel".
                        " path so bailing on this feed.");
                    continue;
                } else {
                    L\crawlLog("...Channel scraped.");
                }
                $nodes = $this->getTags($sub_dom[0], $feed['ITEM_PATH']);
                $rss_elements = ["title" => $feed['TITLE_PATH'],
                    "description" => $feed['DESCRIPTION_PATH'],
                    "link" => $feed['LINK_PATH']];
            } else {
                $nodes = $dom->getElementsByTagName('item');
                // see above comment on why slink rather than link
                $rss_elements = ["title" => "title",
                    "description" => "description", "link" =>"slink",
                    "guid" => "guid", "pubdate" => "pubdate"];
                if ($nodes->length == 0) {
                    // maybe we're dealing with atom rather than rss
                    $nodes = $dom->getElementsByTagName('entry');
                    $rss_elements = [
                        "title" => "title", "description" => "summary",
                        "link" => "slink", "guid" => "id",
                        "pubdate" => "updated"];
                }
            }
            L\crawlLog("...done extracting info. Check for new news ".
                "items in {$feed['NAME']}.");
            $num_added = 0;
            $num_seen = 0;
            foreach ($nodes as $node) {
                $item = [];
                foreach ($rss_elements as $db_element => $feed_element) {
                    L\crawlTimeoutLog("..still adding feed items to index.");
                    if ($is_html) {
                        $tag_nodes = $this->getTags($node, $feed_element);
                        if (!isset($tag_nodes[0])) {
                            $tag_node = null;
                        } else {
                            $tag_node = $tag_nodes[0];
                        }
                        $element_text = (is_object($tag_node)) ?
                            $tag_node->textContent: "";
                    } else {
                        $tag_node = $node->getElementsByTagName(
                                $feed_element)->item(0);
                        $element_text = (is_object($tag_node)) ?
                            $tag_node->nodeValue: "";
                    }
                    if ($db_element == "link" && $tag_node &&
                        ($element_text == "" || $is_html)) {
                        if ($is_html) {
                            $element_text =
                               $tag_node->documentElement->getAttribute("href");
                        } else {
                            $element_text = $tag_node->getAttribute("href");
                        }
                        $element_text = UrlParser::canonicalLink($element_text,
                            $feed["SOURCE_URL"]);
                    }
                    $item[$db_element] = strip_tags($element_text);
                }
                $item['image_link'] = "";
                if (isset($feed['IMAGE_XPATH']) && $feed['IMAGE_XPATH'] != "") {
                    if ($feed['IMAGE_XPATH'][0]=="^") {
                        $is_channel_image = true;
                        $image_xpath = substr($feed['IMAGE_XPATH'], 1);
                    } else {
                        $is_channel_image = false;
                        $image_xpath = $feed['IMAGE_XPATH'];
                    }
                    if ($is_html) {
                        if ($is_channel_image) {
                            $dom_xpath = new \DOMXPath($sub_dom);
                        } else {
                            $dom_xpath = new \DOMXPath($node);
                        }
                        $image_nodes =
                            $dom_xpath->evaluate($image_xpath);
                        if ($image_nodes && $image_nodes->item(0)) {
                            $item['image_link'] =
                                $image_nodes->item(0)->nodeValue;
                        }
                    } else {
                        $dom_xpath = new \DOMXPath($dom);
                        if ($is_channel_image) {
                            $query = $image_xpath;
                        } else {
                            $query = $node->getNodePath() . $image_xpath;
                        }
                        libxml_clear_errors();
                        $image_nodes = $dom_xpath->evaluate($query);
                        if ($image_nodes && $image_nodes->item(0)) {
                            $item['image_link'] =
                                $image_nodes->item(0)->nodeValue;
                        }
                    }
                }
                $did_add = $this->addFeedItemIfNew($item, $feed['NAME'], $lang,
                    $age);
                if ($did_add) {
                    $num_added++;
                }
                $num_seen++;
            }
            L\crawlLog("...added $num_added news items of $num_seen ".
                "on rss page.\n Done Processing {$feed['NAME']}.");
        }
    }
    /**
     * Returns an array of DOMDocuments for the nodes that match an xpath
     * query on $dom, a DOMDocument
     *
     * @param DOMDocument $dom document to run xpath query on
     * @param string $query xpath query to run
     * @return array of DOMDocuments one for each node matching the
     *  xpath query in the orginal DOMDocument
     */
    public function getTags($dom, $query)
    {
        $nodes = [];
        $dom_xpath = new \DOMXPath($dom);
        if (!$dom_xpath) {
            return [];
        }
        $tags = $dom_xpath->query($query);
        if (!$tags) {
            return [];
        }
        $i = 0;
        while ($item = $tags->item($i)) {
            $tmp_dom = new \DOMDocument;
            $tmp_dom->formatOutput = true;
            $node = $tmp_dom->importNode($item, true);
            $tmp_dom->appendChild($node);
            $nodes[] = $tmp_dom;
            $i++;
        }
        return $nodes;
    }
    /**
     * Copies all feeds items newer than $age to a new shard, then deletes
     * old index shard and database entries older than $age. Finally sets copied
     * shard to be active. If this method is going to take max_execution_time/2
     * it returns false, so an additional job can be schedules; otherwise
     * it returns true
     *
     * @param int $age how many seconds old records should be deleted
     * @return bool whether job executed to complete
     */
    public function rebuildFeedShard($age)
    {
        $time = time();
        $feed_shard_name = C\WORK_DIRECTORY."/feeds/index";
        $prune_shard_name = C\WORK_DIRECTORY."/feeds/prune_index";
        $prune_shard =  new IndexShard($prune_shard_name);
        $too_old = $time - $age;
        if (!$prune_shard) {
            return false;
        }
        $pre_feeds = $this->getNewsSources();
        if (!$pre_feeds) {
            return false;
        }
        $feeds = [];
        foreach ($pre_feeds as $pre_feed) {
            if (!isset($pre_feed['NAME'])) {
                continue;
            }
            $feeds[$pre_feed['NAME']] = $pre_feed;
        }
        $db = $this->db;
        // we now rebuild the inverted index with the remaining items
        $sql = "SELECT * FROM FEED_ITEM ".
            "WHERE PUBDATE >= ? ".
            "ORDER BY PUBDATE DESC";
        $result = $db->execute($sql, [$too_old]);
        if ($result) {
            $completed = true;
            L\crawlLog("..still deleting. Making new index".
                " of non-pruned items.");
            $i = 0;
            while ($item = $db->fetchArray($result)) {
                L\crawlTimeoutLog("..have added %s non-pruned items to index.",
                    $i);
                $i++;
                if (!isset($item['SOURCE_NAME'])) {
                    continue;
                }
                $source_name = $item['SOURCE_NAME'];
                if (isset($feeds[$source_name])) {
                    $lang = $feeds[$source_name]['LANGUAGE'];
                } else {
                    $lang = "";
                }
                $phrase_string = $item["TITLE"] . " ". $item["DESCRIPTION"];
                $word_lists = PhraseParser::extractPhrasesInLists(
                    $phrase_string, $lang);
                $raw_guid = L\unbase64Hash($item["GUID"]);
                $doc_keys = L\crawlHash($item["LINK"], true) .
                    $raw_guid."d". substr(L\crawlHash(
                    UrlParser::getHost($item["LINK"])."/", true), 1);
                $meta_ids = $this->calculateMetas($lang, $item['PUBDATE'],
                    $source_name, $item["GUID"]);
                $prune_shard->addDocumentWords($doc_keys, $item['PUBDATE'],
                    $word_lists, $meta_ids, PhraseParser::$materialized_metas,
                    true, false);
            }
        }
        $prune_shard->save();
        @chmod($prune_shard_name, 0777);
        @chmod($feed_shard_name, 0777);
        @rename($prune_shard_name, $feed_shard_name);
        @chmod($feed_shard_name, 0777);
        $sql = "DELETE FROM FEED_ITEM WHERE PUBDATE < ?";
        $db->execute($sql, [$too_old]);
    }
    /**
     * Adds $item to FEED_ITEM table in db if it isn't already there
     *
     * @param array $item data from a single news feed item
     * @param string $source_name string name of the news feed $item was found
     * on
     * @param string $lang locale-tag of the news feed
     * @param int $age how many seconds old records should be ignored
     * @return bool whether an item was added
     */
    public function addFeedItemIfNew($item, $source_name, $lang, $age)
    {
        if (!isset($item["link"]) || !isset($item["title"]) ||
            !isset($item["description"])) {
            return false;
        }
        if (!isset($item["guid"]) || $item["guid"] == "") {
            $item["guid"] = L\crawlHash($item["link"]);
        } else {
            $item["guid"] = L\crawlHash($item["guid"]);
        }
        if (!isset($item["image_link"])) {
            $item["image_link"] = "";
        }
        $raw_guid = L\unbase64Hash($item["guid"]);
        if (!isset($item["pubdate"]) || $item["pubdate"] == "") {
            $item["pubdate"] = time();
        } else {
            $item["pubdate"] = strtotime($item["pubdate"]);
            if ($item["pubdate"] < 0) {
                $item["pubdate"] = time();
            }
        }
        if (time() - $item["pubdate"] > $age) {
            return false;
        }
        $sql = "SELECT COUNT(*) AS NUMBER FROM FEED_ITEM WHERE GUID = ?";
        $db = $this->db;
        $result = $db->execute($sql, [$item["guid"]]);
        if ($result) {
            $row = $db->fetchArray($result);
            if ($row["NUMBER"] > 0) {
                return false;
            }
        } else {
            return true;
        }
        $sql = "INSERT INTO FEED_ITEM (GUID, TITLE, LINK, IMAGE_LINK,".
            "DESCRIPTION, PUBDATE, SOURCE_NAME) VALUES (?, ?, ?, ?, ?, ?, ?)";
        $result = $db->execute($sql, [$item['guid'],
            $item['title'], $item['link'], $item['image_link'],
            $item['description'], $item['pubdate'], $source_name]);
        if (!$result) {
            return false;
        }
        return true;
    }
    /**
     * Used to calculate the meta words for RSS feed items
     *
     * @param string $lang the locale_tag of the feed item
     * @param int $pubdate UNIX timestamp publication date of item
     * @param string $source_name the name of the news feed
     * @param string $guid the guid of the news item
     *
     * @return array $meta_ids meta words found
     */
    public function calculateMetas($lang, $pubdate, $source_name, $guid)
    {
        $meta_ids = ["media:news", "media:news:".urlencode($source_name),
            "guid:".strtolower($guid)];
        $meta_ids[] = 'date:'.date('Y', $pubdate);
        $meta_ids[] = 'date:'.date('Y-m', $pubdate);
        $meta_ids[] = 'date:'.date('Y-m-d', $pubdate);
        $meta_ids[] = 'date:'.date('Y-m-d-H', $pubdate);
        $meta_ids[] = 'date:'.date('Y-m-d-H-i', $pubdate);
        $meta_ids[] = 'date:'.date('Y-m-d-H-i-s', $pubdate);
        if ($lang != "") {
            $lang_parts = explode("-", $lang);
            $meta_ids[] = 'lang:'.$lang_parts[0];
            if (isset($lang_parts[1])) {
                $meta_ids[] = 'lang:'.$lang;
            }
        }
        return $meta_ids;
    }
    /**
     * Used to make an exec call to the crawl controller
     * to get the next available file for conversion
     *
     * @return folder and file name
     */
    public function requestFileForConversion()
    {
        $current_machine = $this->getCurrentMachine();
        $folder_file = $this->execMachines("getSplitFileForConversion",
            [C\NAME_SERVER], $current_machine);
        $value = L\webdecode($folder_file[0][CrawlConstants::PAGE]);
        return $value;
    }
    /**
     * Used to make an exec call to the crawl controller
     * to get the next available file for sending emails
     *
     * @return string file name
     */
    public function requestFileForMailingList()
    {
        $mailer_file = $this->execMachines("getMailingListFile",
            [C\NAME_SERVER]);
        $file_name = L\webdecode($mailer_file[0][CrawlConstants::PAGE]);
        return $file_name;
    }
}
